#!/bin/bash
set -e
srcdir=$(pwd)
pkgname="yong"
buildjs="build.js/build.js"
pkgdir=$(pwd)/debian/tmp

ENABLE_GTK2=1
ENABLE_GTK3=1
ENABLE_QT5=1

BUILD_ARCH=$(dpkg-architecture -qDEB_BUILD_ARCH)
case ${BUILD_ARCH} in
    amd64)
        arch=l64
        ;;
    i*86)
        arch=l32
        ;;
esac

build(){
    mkdir -p $srcdir/$pkgname/{llib,cloud,gbk,mb,vim}/$arch
    if [ ${ENABLE_GTK2} = "1" ];then
        mkdir -p $srcdir/$pkgname/{im,config}/${arch}-gtk2
        mkdir -p $srcdir/$pkgname/im/gtk-im/${arch}-gtk2
    fi

    if [ ${ENABLE_GTK3} =  "1" ];then
        mkdir -p $srcdir/$pkgname/{im,config}/${arch}-gtk3
        mkdir -p $srcdir/$pkgname/im/gtk-im/${arch}-gtk3
    fi
    mkdir -p $srcdir/$pkgname/im/IMdkit/$arch
    cd "$srcdir/$pkgname/im/IMdkit"
    node "$srcdir/$buildjs" -f "$srcdir/$pkgname/im/IMdkit/build.txt" ${arch} 
    cd "$srcdir/$pkgname/"
    case  ${arch} in
        l64)
            sed -i 's/copy_build("l32");//' $srcdir/$pkgname/install/build.txt
            ;;
        l32)
            sed -i 's/copy_build("l64");//' $srcdir/$pkgname/install/build.txt
            ;;
    esac

    sed -i "s/'config',//" "$srcdir/$pkgname/build.txt"
    sed -i "s/'im',//" "$srcdir/$pkgname/build.txt"

    if [  ${ENABLE_GTK2} != "1" ];then
        sed -i '/-gtk2/d' $srcdir/$pkgname/install/build.txt
    fi
    if [  ${ENABLE_GTK3} != "1" ];then
        sed -i '/-gtk3/d' $srcdir/$pkgname/install/build.txt
    fi

    node "$srcdir/$buildjs" -f "$srcdir/$pkgname/build.txt" ${arch}
    if [ ${ENABLE_GTK2} = "1" ];then
        cd $srcdir/$pkgname/im/
        node "$srcdir/$buildjs" -f "$srcdir/$pkgname/im/build.txt" ${arch}-gtk2
        cd $srcdir/$pkgname/im/gtk-im/
        node "$srcdir/$buildjs" -f "$srcdir/$pkgname/im/gtk-im/build.txt" ${arch}-gtk2
        cd $srcdir/$pkgname/config/
        node "$srcdir/$buildjs" -f "$srcdir/$pkgname/config/build.txt" ${arch}-gtk2
    fi
    if [ ${ENABLE_GTK3} = "1" ];then
        cd $srcdir/$pkgname/im/
        node "$srcdir/$buildjs" -f "$srcdir/$pkgname/im/build.txt" ${arch}-gtk3
        cd $srcdir/$pkgname/im/gtk-im/
        node "$srcdir/$buildjs" -f "$srcdir/$pkgname/im/gtk-im/build.txt" ${arch}-gtk3
        cd $srcdir/$pkgname/config/
        node "$srcdir/$buildjs" -f "$srcdir/$pkgname/config/build.txt" ${arch}-gtk3
    fi
    if [ ${ENABLE_QT5} = "1" ];then
        mkdir -p $srcdir/$pkgname/im/qt5-im/${arch}-qt5
        cd $srcdir/$pkgname/im/qt5-im
        node "$srcdir/$buildjs" -f "$srcdir/$pkgname/im/qt5-im/build.txt" ${arch}-qt5

    fi
    cd "$srcdir/$pkgname/install/"
    node "$srcdir/$buildjs" -f "$srcdir/$pkgname/install/build.txt" copy

}

package(){
    case $arch in
        l32)
            multi_arch="i386-linux-gnu"
            ;;
        l64)
            multi_arch="x86_64-linux-gnu"
            ;;
    esac
    mkdir -p $pkgdir/usr/bin
    mkdir -p $pkgdir/usr/share
    cp -a $srcdir/$pkgname/install/yong $pkgdir/usr/share
    cd $pkgdir/usr/share/yong
    install -D locale/zh_CN.mo $pkgdir/usr/share/locale/zh_CN/LC_MESSAGES/yong.mo
    if [ ${ENABLE_GTK2} = "1" ];then
        install -D ${arch}/gtk-im/im-yong-gtk2.so $pkgdir/usr/lib/${multi_arch}/gtk-2.0/2.10.0/immodules/im-yong.so
    fi
    if [ ${ENABLE_GTK3} = "1" ];then
        install -D ${arch}/gtk-im/im-yong-gtk3.so $pkgdir/usr/lib/${multi_arch}/gtk-3.0/3.0.0/immodules/im-yong.so
    fi
    if [ ${ENABLE_QT5} = "1" ];then
        install -D $srcdir/$pkgname/im/qt5-im/${arch}-qt5/libyongplatforminputcontextplugin.so $pkgdir/usr/lib/${multi_arch}/qt5/plugins/platforminputcontexts/libyongplatforminputcontextplugin.so
    fi
    # Fix ${arch} name to runtime
    mv ${arch} runtime
    rm -rf runtime/gtk-im
}

case $1 in
    "build")
        build
        ;;
    "package")
        package
        ;;
esac
